function output = osimOptimalTrackingEndpoint(input)

% Variables at Start and Terminus of Phase 1
t0 = input.phase(1).initialtime;
tf = input.phase(1).finaltime;
x0 = input.phase(1).initialstate;
xf = input.phase(1).finalstate;

integral = input.phase(1).integral;

input.phase(1).state = [x0;xf];
input.phase(1).time = [t0;tf];
[grfs,handForces,residuals] = getGrfs(input);

% Constraints
if input.auxdata.handLoadControls
    output.eventgroup(1).event = ([x0(1:62)-xf(1:62) diff(grfs) residuals]-...
        (input.auxdata.eventMax+input.auxdata.eventMin)/2)...
        ./(input.auxdata.eventMax-input.auxdata.eventMin);
else
    output.eventgroup.event = ([x0([7:31 38:62])-xf([7:31 38:62])  diff(grfs) diff(handForces) residuals]-...
        (input.auxdata.eventMax+input.auxdata.eventMin)/2)...
        ./(input.auxdata.eventMax-input.auxdata.eventMin);
end


% Objective
if input.auxdata.handLoadControls
    q = 1e-2*(sum(integral([13,20,21:31]))*0 ... %13,20,21:23,25,29
        +sum(integral(32:37).*[0,0.1,0.1,0,0.1,0.1])*0 ...
        +sum(integral(38:43).*[0.2,0.2,0,0.2,0.2,0])*0 ...
        +sum(integral(44:55))*0 ...
        +sum(integral(56:65))*0 ...
        +sum(integral(66:end))*0.001); %0.001
else
    q = 1e-3*(sum(integral([1:3 9 13 16 20:31]))*1 ...
        +sum(integral(32:37))*1 ...
        +sum(integral(38:43))*1 ...
        +sum(integral([44:45 47:51 53:55]))*1 ...
        +sum(integral(56:127))*40 ... %40
        +sum(integral(128:139))*1 ...
        +sum(integral(140:end))*40); %40
end

output.objective = q;
end


%% Function Definition
function [grfs, handForces,residuals] = getGrfs(input)

% rescale everything back to original
nPts = length(input.phase.time);
onesCol = ones(nPts,1);

input.phase.time = (input.phase.time)*(input.auxdata.tMax-input.auxdata.tMin)+(input.auxdata.tMax+input.auxdata.tMin)/2;
input.phase.state = (input.phase.state).*(onesCol*(input.auxdata.stateMax-input.auxdata.stateMin))+onesCol*(input.auxdata.stateMax+input.auxdata.stateMin)/2;


%% Load parameters from the auxilary data
tScale = input.auxdata.tScale;

mu_s = 0; % Coefficient of static friction for foot-ground contact
mu_d = 1;%input.phase.parameter(1,9);%input.auxdata.mu;%parameters(1,1); % Coefficient of dynamic friction
mu_v = 0; % Coefficient of viscous friction
kVal = input.auxdata.kVal;%input.auxdata.Kval; % Spring stiffnesses for foot ground contact
cVal = input.auxdata.cVal; % Spring damping
yVal = input.auxdata.yVal;
springMat = input.auxdata.springMat;
springBodyMat = input.auxdata.springBodyMat;
numSpringsBody = input.auxdata.numSpringsBody;

% SpringMat(7:end,2) = SpringMat(7:end,2)+Yval(1);
springMat(7:6+input.auxdata.numSprings/2,2) = springMat(7:6+input.auxdata.numSprings/2,2)+yVal(1);
springMat(7+input.auxdata.numSprings/2:end,2) = springMat(7+input.auxdata.numSprings/2:end,2)+yVal(2);

% Transition velocity between static and dynamic friction
latchvel = input.auxdata.latchvel;


%% Initialize Variables

t = input.phase(1).time;

if ~exist('mkrData','var')
    mcr = fnval(input.auxdata.spMcr,t)';
    mcl = fnval(input.auxdata.spMcl,t)';
elseif size(mkrData,1) ~= nPts
    mcr = fnval(input.auxdata.spMcr,t)';
    mcl = fnval(input.auxdata.spMcl,t)';
end

%% Kinematics

% Set states
x = input.phase(1).state(:,1:31);
xp = input.phase(1).state(:,32:62);

% Calculate spring positions and velocities
[springPos, springVels] = opensimPointKin_v3(t,[x xp],springMat',springBodyMat);

% Store location of heel and toes body for tranferring GRFs
BodyPositions = springPos(:,1:12);
% mcr = SpringPos(:,13:15);
% mcl = SpringPos(:,16:18);
springPos(:,1:18) = [];
springVels(:,1:18) = [];

mcr(:,2) = 0;
mcl(:,2) = 0;

%% Calculate Ground Reaction Forces

% Calculate spring forces and moments about electrical centers
[footGRFsRMC, footGRFsLMC] = calcGroundReactions_v3(springPos, springVels, ...
    numSpringsBody, mu_s, mu_d, mu_v, kVal, cVal, mcr, mcl, latchvel, tScale,input.auxdata.beltSpeed);

% Transfer moments to heel and toe origin for input into full body model
% This is necessary because the contact model ground reactions are
% calculated about some point in the room (force plate electrical centers),  
% and the model applies the loads at the origin of the foot and toes body
footGRFsR = footGRFsRMC;
footGRFsL = footGRFsLMC;

% Transfer to heel
footGRFsR(:,7:9) = cross(mcr-BodyPositions(:,1:3),footGRFsRMC(:,1:3),2)+footGRFsRMC(:,7:9);
footGRFsL(:,7:9) = cross(mcl-BodyPositions(:,7:9),footGRFsLMC(:,1:3),2)+footGRFsLMC(:,7:9);

% Transfer to toe
footGRFsR(:,10:12) = cross(mcr-BodyPositions(:,4:6),footGRFsRMC(:,4:6),2)+footGRFsRMC(:,10:12);
footGRFsL(:,10:12) = cross(mcl-BodyPositions(:,10:12),footGRFsLMC(:,4:6),2)+footGRFsLMC(:,10:12);

%% Calculate Reaction Loads on the Hands
handLoadsR = zeros(nPts,6);
handLoadsL = zeros(nPts,6);

if ~input.auxdata.handLoadControls
    % 3 linear springs, 3 torsion springs per hand
    % need linear displacement and angular displacement

    handSpring_k = input.auxdata.handSpring_k;
    handSpring_tk = input.auxdata.handSpring_tk;
    if input.auxdata.springPosParams
        medianHandPos_r = input.auxdata.medianHandPos_r + input.auxdata.handPosCorrect_r;
        medianHandAng_r = input.auxdata.medianHandAng_r + input.auxdata.handAngCorrect_r;
        medianHandPos_l = input.auxdata.medianHandPos_l + input.auxdata.handPosCorrect_l;
        medianHandAng_l = input.auxdata.medianHandAng_l + input.auxdata.handAngCorrect_l;
    else
        medianHandPos_r = input.auxdata.medianHandPos_r;
        medianHandAng_r = input.auxdata.medianHandAng_r;
        medianHandPos_l = input.auxdata.medianHandPos_l;
        medianHandAng_l = input.auxdata.medianHandAng_l;
    end
    
    handSpring_c = 1e-4;
    handSpring_tc = 1e-5;
    
    handBodyMat = input.auxdata.handBodyMat;
    handForceLoc = input.auxdata.handForceLoc;
    [handPos, handVel, handAng, handAngVel] = opensimGetBodyPosVel_OMP(t,[x xp],handForceLoc,handBodyMat);
    handDisplacement_r = [-handPos(:,1) + medianHandPos_r(1),-handPos(:,2) + medianHandPos_r(2),-handPos(:,3) + medianHandPos_r(3)];
    handDisplacement_l = [-handPos(:,4) + medianHandPos_l(1),-handPos(:,5) + medianHandPos_l(2),-handPos(:,6) + medianHandPos_l(3)];
    handAngDisp_r = [-handAng(:,1) + medianHandAng_r(1),-handAng(:,2) + medianHandAng_r(2),-handAng(:,3) + medianHandAng_r(3)];
    handAngDisp_l = [-handAng(:,4) + medianHandAng_l(1),-handAng(:,5) + medianHandAng_l(2),-handAng(:,6) + medianHandAng_l(3)];


    handLoadsR(:,1) = (handDisplacement_r(:,1)*handSpring_k(1)).*(1+handSpring_c*handVel(:,1)); %TODO: Check this
    handLoadsR(:,2) = (handDisplacement_r(:,2)*handSpring_k(2)).*(1+handSpring_c*handVel(:,2));
    handLoadsR(:,3) = (handDisplacement_r(:,3)*handSpring_k(3)).*(1+handSpring_c*handVel(:,3));
    handLoadsR(:,4) = (handAngDisp_r(:,1)*handSpring_tk(1)).*(1+handSpring_tc*handAngVel(:,1));
    handLoadsR(:,5) = (handAngDisp_r(:,2)*handSpring_tk(2)).*(1+handSpring_tc*handAngVel(:,2));
    handLoadsR(:,6) = (handAngDisp_r(:,3)*handSpring_tk(3)).*(1+handSpring_tc*handAngVel(:,3));

    handLoadsL(:,1) = handDisplacement_l(:,1)*handSpring_k(1).*(1+handSpring_c*handVel(:,4)); %TODO: Check this
    handLoadsL(:,2) = handDisplacement_l(:,2)*handSpring_k(2).*(1+handSpring_c*handVel(:,5));
    handLoadsL(:,3) = handDisplacement_l(:,3)*handSpring_k(3).*(1+handSpring_c*handVel(:,6));
    handLoadsL(:,4) = (handAngDisp_l(:,1)*handSpring_tk(1)).*(1+handSpring_tc*handAngVel(:,4));
    handLoadsL(:,5) = (handAngDisp_l(:,2)*handSpring_tk(2)).*(1+handSpring_tc*handAngVel(:,5));
    handLoadsL(:,6) = (handAngDisp_l(:,3)*handSpring_tk(3)).*(1+handSpring_tc*handAngVel(:,6));
    
end

%% Dynamics

% Extract for input into full body dynamic model
torques = zeros(nPts,25);%Set to zero because they will be determined by inverse dynamics

% Formulate inputs into opensim (states and controls w/ GRFs)
appliedLoads = [torques footGRFsR footGRFsL handLoadsR handLoadsL];

% Set accelerations
xpp = input.phase.state(:,63:end);

% Calculate state derivatives and marker locations
[residualLoads] = opensimID_OMP(t,[x xp], xpp, appliedLoads,[1 0 0]);
residuals = residualLoads(end,1:6);


%% Transfer GRFs

% Calculate net GRFs at the electrical center for use in objective function
footGroundReactionsR = footGRFsR(:,[1:3 7:9])+footGRFsR(:,[4:6 10:12]);
footGroundReactionsL = footGRFsL(:,[1:3 7:9])+footGRFsL(:,[4:6 10:12]);

grfs = [footGroundReactionsR(:,1:3) footGroundReactionsL(:,1:3)];
handForces = [handLoadsR(:,1:3),handLoadsL(:,1:3)];
end